/*
** Copyright 2001, Michael Pfeiffer	
*/

#include <KernelExport.h>
#include <Drivers.h>
#include <Errors.h>
#include <FindDirectory.h>
#include <image.h>
#include <stdio.h>
#include <string.h>
#include <graphic_driver.h>
#include <unistd.h>
#include "generic.h"
#include "../include/DriverInterface.h"
#include <be/drivers/driver_settings.h>

#ifdef DEBUG
static int log_fd = -1;
#define LOG(s)    { if (log_fd != -1) write(log_fd, s, strlen(s)); }
#define START_LOG { log_fd = open("/boot/var/log/wrapper.accelerant", O_RDWR | O_CREAT | O_TEXT | O_TRUNC, 0333);}
#define STOP_LOG  { close(log_fd); log_fd = -1; }
#else
#define LOG(s)    {}
#define START_LOG {}
#define STOP_LOG  {}
#endif

// area, so wrapper accelerant can retrieve the name of the wrapped accelerant 
static int                initialized = false;
static int                driver_fd = -1;
static struct SharedData *shared = NULL;
static area_id            shared_id = -1;
static image_id           accl_id = -1;

static void *   (*a_get_accelerant_hook)(uint32 feature, void *data) = NULL;
static status_t (*A_INIT_ACCELERANT)(int fd) = NULL;
static void     (*A_UNINIT_ACCELERANT)(void) = NULL;
static status_t (*A_SET_DISPLAY_MODE)(display_mode *mode_to_set) = NULL;
static status_t (*A_CLONE_ACCELERANT)(void *data);

/*

The standard entry point.  Given a uint32 feature identifier, this routine
returns a pointer to the function that implements the feature.  Some features
require more information than just the identifier to select the proper
function.  The extra information (which is specific to the feature) is
pointed at by the void *data parameter.  By default, no extra information
is available.  Any extra information available to choose the function will be
noted on a case by case below.

*/

#define WRAP(x) case B_##x: LOG(#x); return (void *)x
status_t init(uint32 feature, void* data, bool clone);
void * _EXPORT get_accelerant_hook(uint32 feature, void *data) {
#ifdef DEBUG
	char buffer[80];
	LOG("accelerant: get_accelerant_hook ");
	sprintf(buffer, "%d %x\n", (int)feature, (int)data);
	LOG(buffer);
#endif

	switch (feature) {
	case B_INIT_ACCELERANT:
		init(feature, data, false); return (void*)INIT_ACCELERANT;
		break;
	case B_CLONE_ACCELERANT:
		init(feature, data, true); return (void*)CLONE_ACCELERANT;
		break;
	case B_UNINIT_ACCELERANT:
		A_UNINIT_ACCELERANT = a_get_accelerant_hook(feature, data);
		return (void*)UNINIT_ACCELERANT;
		break;
	case B_SET_DISPLAY_MODE:
		A_SET_DISPLAY_MODE = a_get_accelerant_hook(feature, data);
		return (void*)SET_DISPLAY_MODE;
		break;
	default:
		/* Delegate call to wrapped accelerant */
		if (a_get_accelerant_hook) return a_get_accelerant_hook(feature, data);
		LOG("accelerant: a_get_accelerant_hook is NULL\n");
		return 0;
	}
}

#undef WRAP


bool load_accelerant(directory_which which, char* name) {
	char path[B_PATH_NAME_LENGTH+1];
	if (B_OK != find_directory(which, 0, false, path, B_PATH_NAME_LENGTH)) { accl_id = -1; return false; }
	strcat(path, "/accelerants/");
	strcat(path, name);
	LOG(path); LOG("\n");
	accl_id = load_add_on(path);
	return accl_id > 0;
}


void unload_accelerant() {
	if (accl_id != -1) {
		unload_add_on(accl_id); accl_id = -1;
	}
}


status_t get_shared_data() {
	area_id id = find_area(AREA_NAME);
	if (id == B_NAME_NOT_FOUND) return B_ERROR;
	shared_id = clone_area(AREA_NAME, (void**)&shared, B_ANY_ADDRESS, B_READ_AREA | B_WRITE_AREA, id);
	if (shared_id == B_BAD_VALUE || shared_id == B_NO_MEMORY || shared_id == B_ERROR) {
		shared_id = -1; return B_ERROR;
	}
	return B_OK;
}


void release_shared_data() {
	if (shared_id != -1) {
		delete_area(shared_id); shared_id = -1; shared = NULL;
	}
}


status_t init(uint32 feature, void* data, bool clone) {
	status_t err;
	if (initialized) return B_OK; // return, already initialized
	initialized = true;
	START_LOG;
	LOG("accelerant: enter INIT_ACCELERANT\n");
	// get accelerant name from shared area
	err = get_shared_data(); 
	// load accelerant
	if (err == B_OK && 
	    (load_accelerant(B_USER_ADDONS_DIRECTORY, shared->accelerant) ||
	   	load_accelerant(B_BEOS_ADDONS_DIRECTORY, shared->accelerant))) {
	   	
	   	err = get_image_symbol(accl_id, "get_accelerant_hook", B_SYMBOL_TYPE_TEXT, (void**)&a_get_accelerant_hook);

	    if (err != B_OK) {
	    	a_get_accelerant_hook = NULL; 
		} else {
			if (clone) {
				A_CLONE_ACCELERANT = a_get_accelerant_hook(B_CLONE_ACCELERANT, data);
			} else {
				A_INIT_ACCELERANT = a_get_accelerant_hook(B_INIT_ACCELERANT, data);
			}
		}
	} else {
		err = B_ERROR;
	}
	
	if (err != B_OK) {
		unload_accelerant();
		release_shared_data();
		STOP_LOG;
	} else {
		LOG("leave INIT_ACCELERANT\n");
	}
	
	return err;
}


status_t INIT_ACCELERANT(int fd) {
	driver_fd = fd;
	if (A_INIT_ACCELERANT) {
		return A_INIT_ACCELERANT(fd);
	} 
	return B_ERROR;
}


status_t CLONE_ACCELERANT(void* data) {
	LOG("accelerant: enter CLONE_ACCELERANT\n");
	if (A_CLONE_ACCELERANT) return A_CLONE_ACCELERANT(data);
	return B_ERROR;
}


void UNINIT_ACCELERANT(void) {
	LOG("accelerant: enter UNINIT_ACCELERANT\n");
	if (A_UNINIT_ACCELERANT) A_UNINIT_ACCELERANT();
	unload_accelerant();
	release_shared_data();
	STOP_LOG;
}

static struct {
	int width;
	int height;
} modes[] = {
	{  640, 480},
	{  800, 600},
	{ 1024, 768},
	{ 1152, 768},
	{ 1280, 1024},
	{ 1600, 1200}
};

static status_t hack_display_mode(display_mode *mode_to_set) {
	status_t err;
	int      i;
	int      rf;
	bool     enabled = ioctl(driver_fd, GDW_IS_ENABLED, NULL, 0) == B_OK;

#ifdef DEBUG
	char buffer[255];

	sprintf(buffer, "%dx%d/%dHz\n", mode_to_set->timing.h_display, mode_to_set->timing.v_display,
		(int)(mode_to_set->timing.pixel_clock * 1000 / mode_to_set->timing.h_total / 
		mode_to_set->timing.v_total));
	LOG(buffer);
#endif

	if ((!enabled) || mode_to_set == NULL) return B_ERROR;
	
	// search full screen display mode
	for (i = 0; i < 6; i++) {
		if (mode_to_set->timing.h_display == modes[i].width &&
		    mode_to_set->timing.v_display == modes[i].height) break;
	}
	
	// calculate current refresh rate
	rf = mode_to_set->timing.pixel_clock * 1000 / mode_to_set->timing.h_total / 
		mode_to_set->timing.v_total;	
	
	
	if (i < 6 && 59 <= rf && rf <= 61) {
		// change mode	
		display_mode mode = *mode_to_set;
		int32 hz = i;
		// get refresh rate for mode "i"
		if (B_OK != ioctl(driver_fd, GDW_FULLSCREEN_REFRESHRATE, &hz, sizeof(int32))) {
#ifdef DEBUG
			LOG("GDW_FULLSCREEN_REFRESHRATE return error\n");
			sprintf(buffer, "hz = %d\n", (int)hz);
			LOG(buffer);
#endif
			hz = 60;
		}
		if (hz >= 60 && hz <= 100) {
			// change pixel clock
			mode.timing.pixel_clock = mode.timing.h_total * mode.timing.v_total * hz / 1000;
			err = A_SET_DISPLAY_MODE(&mode);
			if (err == B_OK) { 
				LOG("fullscreen mode set\n"); 
				/**mode_to_set = mode;*/ 
				return B_OK; 
			} else {
				LOG("error could not set fullscreen mode\n");
			}
		} else {
#ifdef DEBUG
			sprintf(buffer, "Unsupported refreshrate %d Hz\n", (int)hz);
			LOG(buffer);
#endif
		}
	}
	return B_ERROR; 
}

status_t SET_DISPLAY_MODE(display_mode *mode_to_set) {
	status_t err;
	
	LOG("accelerant: enter SET_DISPLAY_MODE\n");
	// ASSERT(A_SET_DISPLAY_MODE != NULL)
	
	err = hack_display_mode(mode_to_set);
	if (err == B_OK) return B_OK;
	
	// delegate call to wrapped accelerant
	err = A_SET_DISPLAY_MODE(mode_to_set);
	return err;
}


